//
//  Validate.swift
//  OCR-SDK
//
//  Created by itsol on 6/10/20.
//  Copyright © 2020 itsol. All rights reserved.
//

import Foundation
import UIKit
import Vision

class SBKValidateInput {
    static let shared = SBKValidateInput()

    var modelDataHandler: SBKModelDataHandler? = SBKModelDataHandler(modelFileInfo: MobileNet.cardModel)
    var modelDataFaceHandler: SBKModelDataHandler? = SBKModelDataHandler(modelFileInfo: MobileNet.modelInfo)
    public typealias CompletionHandle = (_ data: Bool) -> Void
    
    func validateCard(imageInput: UIImage) -> Int {
        let ciimage = CIImage(image: imageInput)
        
        let tmpcontext = CIContext(options: nil)
        let cgimage =  tmpcontext.createCGImage(ciimage!, from: ciimage!.extent)
        
        let cvpixelImge = self.convertCGImgeToCVPixelBuffer(forImage: cgimage!)
        let resultVaidate = self.didOutput(pixelBuffer: cvpixelImge!)
        
        switch resultVaidate {
        case .ERROR:
            return -1
        case .IMAGE_FAKE:
            return 0
        case .IMAGE_FRONT:
            return 1
        case .IMAGE_BACK:
            return 2
        }
    }
    
    func validateFace(imageFace: CVPixelBuffer, completion: @escaping CompletionHandle) {
//        let ciimage = CIImage(image: imageFace)
//
//        let tmpcontext = CIContext(options: nil)
//        let cgimage =  tmpcontext.createCGImage(ciimage!, from: ciimage!.extent)
//
//        let cvpixelImge = self.convertCGImgeToCVPixelBuffer(forImage: cgimage!)
        self.detectFace(in: imageFace) { [weak self] data in
            guard let `self` = self else { return }
            let result = data && self.didOutputFace(pixelBuffer: imageFace)
            completion(result)
        }
    }
    
    func didOutput(pixelBuffer: CVPixelBuffer) -> ValidateCard {
        var result: [Float]?
        var previousInferenceTimeMs: TimeInterval = Date.distantPast.timeIntervalSince1970 * 1000
        let delayBetweenInferencesMs: Double = 1000
        
        let currentTimeMs = Date().timeIntervalSince1970 * 1000
        guard (currentTimeMs - previousInferenceTimeMs) >= delayBetweenInferencesMs else { return .ERROR }
        previousInferenceTimeMs = currentTimeMs

        // Pass the pixel buffer to TensorFlow Lite to perform inference.
        result = modelDataHandler?.runModel(onFrame: pixelBuffer)
        if result![0] > result![1] &&  result![0] > result![2]{
            return .IMAGE_FAKE
        } else  if result![1] > result![0] &&  result![1] > result![2] {
            return .IMAGE_FRONT
        } else {
            return .IMAGE_BACK
        }
    }
    
    func didOutputFace(pixelBuffer: CVPixelBuffer) -> Bool {
        var result: [Float]?
        var previousInferenceTimeMs: TimeInterval = Date.distantPast.timeIntervalSince1970 * 1000
        let delayBetweenInferencesMs: Double = 1000
        
        let currentTimeMs = Date().timeIntervalSince1970 * 1000
        guard (currentTimeMs - previousInferenceTimeMs) >= delayBetweenInferencesMs else { return false }
        previousInferenceTimeMs = currentTimeMs

        // Pass the pixel buffer to TensorFlow Lite to perform inference.
        result = modelDataFaceHandler?.runModel(onFrame: pixelBuffer)
        if result![0] < result![1] {
            return true
        } else {
            return false
        }
    }
  
    func comvertUIImageToCVPixel(imageInput: UIImage) -> CVPixelBuffer {
        let ciimage = CIImage(image: imageInput)
        let tmpcontext = CIContext(options: nil)
        let cgimage = tmpcontext.createCGImage(ciimage!, from: ciimage!.extent)
        
        return self.convertCGImgeToCVPixelBuffer(forImage: cgimage!)!
    }
    
    func convertCVPixelToUIImage(pixelBuffer: CVPixelBuffer) -> UIImage {
        let ciimage : CIImage = CIImage(cvPixelBuffer: pixelBuffer)
        let imageView : UIImage = self.convertCIToUIImage(cmage: ciimage)
        return imageView
    }
    
    func convertCIToUIImage(cmage: CIImage) -> UIImage {
         let context:CIContext = CIContext.init(options: nil)
         let cgImage:CGImage = context.createCGImage(cmage, from: cmage.extent)!
         let image:UIImage = UIImage.init(cgImage: cgImage)
         return image
    }
    
    func convertCGImgeToCVPixelBuffer (forImage image: CGImage) -> CVPixelBuffer? {
        
        let frameSize = CGSize(width: image.width, height: image.height)
        
        var pixelBuffer:CVPixelBuffer? = nil
        let status = CVPixelBufferCreate(kCFAllocatorDefault, Int(frameSize.width), Int(frameSize.height), kCVPixelFormatType_32BGRA , nil, &pixelBuffer)
        
        if status != kCVReturnSuccess {
            return nil
        }
        
        CVPixelBufferLockBaseAddress(pixelBuffer!, CVPixelBufferLockFlags.init(rawValue: 0))
        let data = CVPixelBufferGetBaseAddress(pixelBuffer!)
        let rgbColorSpace = CGColorSpaceCreateDeviceRGB()
        let bitmapInfo = CGBitmapInfo(rawValue: CGBitmapInfo.byteOrder32Little.rawValue | CGImageAlphaInfo.premultipliedFirst.rawValue)
        let context = CGContext(data: data, width: Int(frameSize.width), height: Int(frameSize.height), bitsPerComponent: 8, bytesPerRow: CVPixelBufferGetBytesPerRow(pixelBuffer!), space: rgbColorSpace, bitmapInfo: bitmapInfo.rawValue)
        
        
        context?.draw(image, in: CGRect(x: 0, y: 0, width: image.width, height: image.height))
        
        CVPixelBufferUnlockBaseAddress(pixelBuffer!, CVPixelBufferLockFlags(rawValue: 0))
        
        return pixelBuffer
    }
    
    func detectFace(in image: CVPixelBuffer, completion: @escaping CompletionHandle) {
        if #available(iOS 11.0, *) {
            let faceDetectionRequest = VNDetectFaceLandmarksRequest(completionHandler: { (request: VNRequest, error: Error?) in
                DispatchQueue.main.async {
                    if (request.results as? [VNFaceObservation]) != nil {
                        if let results = request.results as? [VNFaceObservation], results.count > 0 {
                            completion(true)
                        } else {
                            completion(false)
                        }
                    }
                }
            })
            
            let imageRequestHandler = VNImageRequestHandler(cvPixelBuffer: image, orientation: .leftMirrored, options: [:])
            try? imageRequestHandler.perform([faceDetectionRequest])
        } else {
            // Fallback on earlier versions
        }
    }
    
    //Xử lý ảnh hiển thị
    func cropImage(image: UIImage, rect: CGRect, scale: CGFloat) -> UIImage? {
        let imageCap = image
        
        let widthCrop = imageCap.size.width - imageCap.size.width / 10
        
        UIGraphicsBeginImageContextWithOptions(CGSize(width: widthCrop, height: widthCrop * 3 / 4), true, 0.0)
        image.draw(at: CGPoint(x: -rect.origin.x / scale, y: -rect.origin.y / scale))
        let croppedImage = UIGraphicsGetImageFromCurrentImageContext()
        UIGraphicsEndImageContext()
        return croppedImage
    }
    
    func cropImageHorizontal(image: UIImage, rect: CGRect, scale: CGFloat) -> UIImage? {
        UIGraphicsBeginImageContextWithOptions(CGSize(width: rect.width, height: rect.height), true, 0.0)
        image.draw(at: CGPoint(x: -rect.origin.x / scale, y: -rect.origin.y / scale))
        let croppedImage = UIGraphicsGetImageFromCurrentImageContext()
        UIGraphicsEndImageContext()
        return croppedImage
    }
    
    func cropImageFace(image: UIImage, rect: CGRect, scale: CGFloat) -> UIImage? {
        let imageCap = image
        
        let widthCrop = imageCap.size.width - imageCap.size.width / 4
        
        UIGraphicsBeginImageContextWithOptions(CGSize(width: widthCrop, height: widthCrop * 4 / 3), true, 0.0)
        image.draw(at: CGPoint(x: -rect.origin.x / scale, y: -rect.origin.y / scale))
        let croppedImage = UIGraphicsGetImageFromCurrentImageContext()
        UIGraphicsEndImageContext()
        return croppedImage
    }
}
